
# This file was generated by 'versioneer.py' (0.23) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2023-07-13T12:28:23+0200",
 "dirty": false,
 "error": null,
 "full-revisionid": "6bfbff508b80e588cc03711798b4e2577ad46449",
 "version": "5.6.1"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)
